"""The :mod:`slip` module provides functions and classes
that allow encoding and decoding of messages according to
the SLIP protocol, as defined in :rfc:`1055`.

The module defines encoding and decoding functions,
an incremental encoder and decoder, and stream wrappers
for the slip protocol.
It also registers these with the :mod:`codecs` package.
"""

__version__ = '0.0.1'

__all__ = ['encode', 'decode', 'SlipDecodingError',
           'END', 'ESC', 'ESC_END', 'ESC_ESC', '__version__']

from .slip import (encode, decode, SlipDecodingError,
                   SlipIncrementalDecoder, SlipIncrementalEncoder,
                   SlipStreamReader, SlipStreamWriter,
                   END, ESC, ESC_END, ESC_ESC)


'''
Created on 29 jun. 2015

@author: Ruud de Jong
'''

import codecs
import io
import re
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable
from functools import partial

class SlipDecodingError(ValueError):
    '''Indicates that a byte sequence cannot be decoded.'''


END = 0xc0
ESC = 0xdb
ESC_END = 0xdc
ESC_ESC = 0xdd
"""These constants represent the values for the special SLIP bytes."""

_ENDb = bytes((END,))
_ESCb = bytes((ESC,))
_E_ENDb = bytes((ESC_END,))
_E_ESCb = bytes((ESC_ESC,))
_ESC_ENDb = bytes((ESC, ESC_END))
_ESC_ESCb = bytes((ESC, ESC_ESC))

_invalid_esc_sequence_re = re.compile(_ESCb + b'(?:([^'+_E_ENDb+_E_ESCb
                                      +br'])|(\Z))')

class SlipEncoder():
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.encoded_bytes = bytearray()

    def encode(self, obj, errors=None, final=False):
        if not isinstance(obj, Iterable):
            obj = (obj,)
        self.encoded_bytes.extend(bytes(obj))
        
        if final:
            packet = bytearray()
            if self.encoded_bytes:
                packet.append(END)
                packet.extend(self.encoded_bytes.replace(_ESCb, _ESC_ESCb).
                                                 replace(_ENDb, _ESC_ENDb))
                packet.append(END)
            self.reset()
            return packet

    def reset(self):
        del self.encoded_bytes[:]
    
    def getstate(self):
        return 0
    
    def setstate(self, state):
        pass
        

class SlipDecoder():        
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.input_buffer = b''

    def decode(self, obj, errors=None, final=False):
        if not isinstance(obj, Iterable):
            obj = (obj,)
        self.input_buffer += bytes(obj)
        self.input_buffer = self.input_buffer.lstrip(_ENDb)
        
        packet = b''
        if END in self.input_buffer:
            packet, rest = self.input_buffer.split(_ENDb, 1)
            self.input_buffer = rest.lstrip(_ENDb)
        elif final:
            packet = self.input_buffer
            self.input_buffer = b''
        
        if packet or final:
            # Verify that ESC is always followed by ESC_END or ESC_ESC
            # If not, something has gone wrong in the encoding process
            match = _invalid_esc_sequence_re.search(packet)
            if match:
                b = match.group(1)
                if b:
                    msg = 'Invalid escape sequence ESC-{}'.format(b)
                else:
                    msg = 'Unfinished escape sequence'
                raise SlipDecodingError(msg)                  

            if final and self.input_buffer:
                msg = 'Remaining undecoded bytes: {!r}'.format(self.input_buffer)
                self.reset()
                raise SlipDecodingError(msg)
                
            return packet.replace(_ESC_ENDb, _ENDb).replace(_ESC_ESCb, _ESCb)
        
    def reset(self):
        self.input_buffer = b''
    
    def getstate(self):
        return 0
    
    def setstate(self, state):
        pass
        
        
encode = partial(SlipEncoder().encode, errors=None, final=True)
decode = partial(SlipDecoder().decode, errors=None, final=True)

encode.__doc__ = '''Encode a message in a SLIP packet.

The function returns a SLIP packet with the encoded message.
The SLIP packet contains exactly one
leading and trailing :data:`END` byte.

Args:
    obj (byte or byte iterable): the message to be encoded.

Returns:
    bytes: the SLIP-encoded message.
'''
decode.__doc__ = '''Decode a SLIP packet.

The function must be called with exactly one SLIP packet.
This means that, apart from leading and trailing :data:`END` bytes,
*obj* must contain no other :data:`END` bytes.

Args:
    obj (byte iterable): the SLIP packet to be decoded.

Returns:
    bytes: the message contained in the packet.
    
Raises:
    SlipDecodingError: when *obj* cannot be decoded.
'''


class SlipIncrementalEncoder(SlipEncoder, codecs.IncrementalEncoder):
    pass


class SlipIncrementalDecoder(SlipDecoder, codecs.IncrementalDecoder):
    pass


class SlipStreamWriter(SlipEncoder, codecs.StreamWriter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
    def write(self, msg):
        packet = self.encode(msg, final=True, errors=self.errors)
        self.stream.write(packet)
    
    def writelines(self, msg_list):
        for msg in msg_list:
            self.write(msg)
            

class SlipStreamReader(SlipDecoder, codecs.StreamReader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def read(self, size=None, chars=None, firstline=None):
        final = False
        msg = self.decode(b'', final=final)
        while msg is None:
            buffer = self.stream.read(io.DEFAULT_BUFFER_SIZE)
            if len(buffer) == 0:
                final=True
            msg = self.decode(buffer, final=final)
        return msg
    
    def readline(self, keepends=False):
        return self.read()
    
    def readlines(self, n, keepends=False):
        return [self.read() for _ in range(n)]
        
    
def _search_slip_codec(encoding):
    if encoding == 'slip':
        return codecs.CodecInfo(
                name='Slip',
                encode=encode,
                decode=decode,
                incrementalencoder=SlipIncrementalEncoder,
                incrementaldecoder=SlipIncrementalDecoder,
                streamreader=SlipStreamReader,
                streamwriter=SlipStreamWriter,
        )
    return None

codecs.register(_search_slip_codec)




'''
Created on 1 mrt. 2015

@author: Ruud de Jong
'''

from setuptools import setup, find_packages  # Always prefer setuptools over distutils
from setuptools.command.test import test as TestCommand
import os
import sys

import sliplib

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the relevant file
def read_long_description(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    seperator = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return seperator.join(buf)

class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', 'Arguments to pass to py.test')]
    
    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []
    
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True
    
    def run(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)
    
long_description = read_long_description('README.rst')
  
setup(
    name='sliplib',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version = sliplib.__version__,

    description='Slip package',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/rhjdjong/SlipLib',

    # Author details
    author='Ruud de Jong',
    author_email='ruud.de.jong@xs4all.nl',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
#         'Programming Language :: Python :: 2',
#         'Programming Language :: Python :: 2.6',
#         'Programming Language :: Python :: 2.7',
#         'Programming Language :: Python :: 3',
#         'Programming Language :: Python :: 3.2',
#         'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],

    # What does your project relate to?
    keywords='slip message framing protocol RFC1055',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(),

    # List run-time dependencies here.  These will be installed by pip when your
    # project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    # install_requires=[],

    # List additional groups of dependencies here (e.g. development dependencies).
    # You can install these using the following syntax, for example:
    # $ pip install -e .[dev,test]
    # extras_require = {},

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    # package_data={},

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages.
    # see http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    #data_files=[('my_data', ['data/data_file'])],

    
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    # entry_points={},
    
    tests_require = ['pytest'],
    cmdclass = {'test': PyTest}
)